/*
 Infomaniak kDrive - iOS App
 Copyright (C) 2021 Infomaniak Network SA

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import InfomaniakCore
import kDriveCore
import kDriveResources
import Kingfisher
import UIKit

protocol FileCellDelegate: AnyObject {
    func didTapMoreButton(_ cell: FileCollectionViewCell)
}

@MainActor class FileViewModel {
    static let observedProperties = ["name", "rawType", "_capabilities", "dropbox", "rawVisibility", "extensionType", "isFavorite", "deletedAt", "lastModifiedAt", "isAvailableOffline", "categories", "size", "hasThumbnail", "color"]
    var file: File
    var selectionMode: Bool
    private var downloadProgressObserver: ObservationToken?
    private var downloadObserver: ObservationToken?
    var thumbnailDownloadTask: Kingfisher.DownloadTask?

    var title: String { file.name }

    var icon: UIImage { file.icon }

    var iconTintColor: UIColor? { file.tintColor }

    var iconAccessibilityLabel: String { file.convertedType.title }

    var isFavorite: Bool { file.isFavorite }

    var moreButtonHidden: Bool { selectionMode }

    var categories = [kDriveCore.Category]()

    private var formattedDate: String {
        if let deletedAt = file.deletedAt {
            return Constants.formatFileDeletionRelativeDate(deletedAt)
        } else {
            return Constants.formatFileLastModifiedRelativeDate(file.lastModifiedAt)
        }
    }

    var subtitle: String {
        if let fileSize = file.getFileSize() {
            return fileSize + " • " + formattedDate
        } else {
            return formattedDate
        }
    }

    var isAvailableOffline: Bool {
        file.isAvailableOffline && FileManager.default.fileExists(atPath: file.localUrl.path)
    }

    init(driveFileManager: DriveFileManager, file: File, selectionMode: Bool) {
        self.file = file
        self.selectionMode = selectionMode
        categories = driveFileManager.drive.categories(for: file)
    }

    func setUpDownloadObserver(_ handler: @escaping (Bool, Bool, Double) -> Void) {
        downloadProgressObserver?.cancel()
        downloadProgressObserver = DownloadQueue.instance.observeFileDownloadProgress(self, fileId: file.id) { [weak self] _, progress in
            DispatchQueue.main.async { [weak self] in
                guard let self = self else { return }
                handler(!self.file.isAvailableOffline || progress < 1, progress >= 1 || progress == 0, progress)
            }
        }
        downloadObserver?.cancel()
        downloadObserver = DownloadQueue.instance.observeFileDownloaded(self, fileId: file.id) { [weak self] _, _ in
            DispatchQueue.main.async { [weak self] in
                guard let self = self else { return }
                handler(!self.isAvailableOffline, true, 1)
            }
        }
    }

    func setThumbnail(on imageView: UIImageView) {
        guard (file.convertedType == .image || file.convertedType == .video) && file.hasThumbnail else { return }
        // Configure placeholder
        imageView.image = nil
        imageView.contentMode = .scaleAspectFill
        imageView.layer.cornerRadius = UIConstants.imageCornerRadius
        imageView.layer.masksToBounds = true
        imageView.backgroundColor = KDriveResourcesAsset.loaderDefaultColor.color
        // Fetch thumbnail
        thumbnailDownloadTask = file.getThumbnail { [requestFileId = file.id, weak self] image, _ in
            if self?.file.id == requestFileId {
                imageView.image = image
                imageView.backgroundColor = nil
            }
        }
    }

    deinit {
        downloadProgressObserver?.cancel()
        downloadObserver?.cancel()
        thumbnailDownloadTask?.cancel()
    }
}

class FileCollectionViewCell: UICollectionViewCell, SwipableCell {
    internal var swipeStartPoint: CGPoint = .zero
    internal var initialTrailingConstraintValue: CGFloat = 0

    @IBOutlet weak var disabledView: UIView!
    @IBOutlet weak var contentInsetView: UIView!
    @IBOutlet weak var titleLabel: UILabel!
    @IBOutlet weak var detailLabel: UILabel?
    @IBOutlet weak var logoImage: UIImageView!
    @IBOutlet weak var accessoryImage: UIImageView?
    @IBOutlet weak var moreButton: UIButton!
    @IBOutlet weak var favoriteImageView: UIImageView?
    @IBOutlet weak var collectionView: UICollectionView!
    @IBOutlet weak var availableOfflineImageView: UIImageView!
    @IBOutlet weak var centerTitleConstraint: NSLayoutConstraint!
    @IBOutlet weak var innerViewTrailingConstraint: NSLayoutConstraint!
    @IBOutlet weak var innerViewLeadingConstraint: NSLayoutConstraint!
    @IBOutlet weak var swipeActionsView: UIStackView?
    @IBOutlet weak var stackViewTrailingConstraint: NSLayoutConstraint?
    @IBOutlet weak var detailsStackView: UIStackView?
    @IBOutlet weak var downloadProgressView: RPCircularProgress?
    @IBOutlet weak var highlightedView: UIView!

    private var viewModel: FileViewModel!

    weak var delegate: FileCellDelegate?

    override var isSelected: Bool {
        didSet {
            configureForSelection()
        }
    }

    override var isHighlighted: Bool {
        didSet {
            setHighlighting()
        }
    }

    var checkmarkImage: UIImageView? {
        return logoImage
    }

    static var emptyCheckmarkImage: UIImage = {
        let size = CGSize(width: 24, height: 24)
        let lineWidth = 1.0
        let renderer = UIGraphicsImageRenderer(size: size)

        return renderer.image { ctx in
            ctx.cgContext.setFillColor(KDriveResourcesAsset.backgroundCardViewColor.color.cgColor)
            ctx.cgContext.setStrokeColor(KDriveResourcesAsset.borderColor.color.cgColor)
            ctx.cgContext.setLineWidth(lineWidth)

            let diameter = min(size.width, size.height) - lineWidth
            let rect = CGRect(x: (size.width - diameter) / 2, y: (size.height - diameter) / 2, width: diameter, height: diameter)
            ctx.cgContext.addEllipse(in: rect)
            ctx.cgContext.drawPath(using: .fillStroke)
        }
    }()

    override func awakeFromNib() {
        super.awakeFromNib()
        favoriteImageView?.isAccessibilityElement = true
        favoriteImageView?.accessibilityLabel = KDriveResourcesStrings.Localizable.favoritesTitle
        availableOfflineImageView?.isAccessibilityElement = true
        availableOfflineImageView?.accessibilityLabel = KDriveResourcesStrings.Localizable.buttonAvailableOffline
        moreButton.accessibilityLabel = KDriveResourcesStrings.Localizable.buttonMenu
        downloadProgressView?.setInfomaniakStyle()
        collectionView?.delegate = self
        collectionView?.dataSource = self
        collectionView?.register(cellView: CategoryBadgeCollectionViewCell.self)
    }

    override func prepareForReuse() {
        super.prepareForReuse()
        collectionView?.isHidden = true
        collectionView?.reloadData()
        centerTitleConstraint?.isActive = false
        resetSwipeActions()
        detailLabel?.text = ""
        logoImage.image = nil
        logoImage.backgroundColor = nil
        logoImage.contentMode = .scaleAspectFit
        logoImage.layer.cornerRadius = 0
        logoImage.layer.masksToBounds = false
        detailsStackView?.isHidden = false
        availableOfflineImageView?.isHidden = true
        downloadProgressView?.isHidden = true
        downloadProgressView?.updateProgress(0, animated: false)
        viewModel?.thumbnailDownloadTask?.cancel()
    }

    func initStyle(isFirst: Bool, isLast: Bool) {
        if isLast && isFirst {
            contentInsetView.roundCorners(corners: [.layerMaxXMaxYCorner, .layerMaxXMinYCorner, .layerMinXMaxYCorner, .layerMinXMinYCorner], radius: 10)
        } else if isFirst {
            contentInsetView.roundCorners(corners: [.layerMaxXMinYCorner, .layerMinXMinYCorner], radius: 10)
        } else if isLast {
            contentInsetView.roundCorners(corners: [.layerMaxXMaxYCorner, .layerMinXMaxYCorner], radius: 10)
        } else {
            contentInsetView.roundCorners(corners: [.layerMaxXMaxYCorner, .layerMaxXMinYCorner, .layerMinXMaxYCorner, .layerMinXMinYCorner], radius: 0)
        }
        contentInsetView.clipsToBounds = true
    }

    func setEnabled(_ enabled: Bool) {
        if enabled {
            disabledView.isHidden = true
            disabledView.superview?.sendSubviewToBack(disabledView)
            isUserInteractionEnabled = true
        } else {
            disabledView.backgroundColor = KDriveResourcesAsset.backgroundCardViewColor.color
            disabledView.isHidden = false
            disabledView.superview?.bringSubviewToFront(disabledView)
            isUserInteractionEnabled = false
        }
    }

    func setHighlighting() {
        highlightedView?.isHidden = !isHighlighted
    }

    func configure(with viewModel: FileViewModel) {
        self.viewModel = viewModel
        logoImage.isAccessibilityElement = true
        logoImage.accessibilityLabel = viewModel.iconAccessibilityLabel
        logoImage.image = viewModel.icon
        logoImage.tintColor = viewModel.iconTintColor
        if !viewModel.selectionMode || checkmarkImage != logoImage {
            // In list mode, we don't fetch the thumbnail if we are in selection mode
            viewModel.setThumbnail(on: logoImage)
        }
        titleLabel.text = viewModel.title
        detailLabel?.text = viewModel.subtitle
        favoriteImageView?.isHidden = !viewModel.isFavorite
        availableOfflineImageView.isHidden = !viewModel.isAvailableOffline
        moreButton.isHidden = viewModel.moreButtonHidden
        collectionView?.isHidden = viewModel.categories.isEmpty
        collectionView?.reloadData()

        viewModel.setUpDownloadObserver { [weak self] availableOfflineHidden, downloadProgressHidden, progress in
            self?.availableOfflineImageView?.isHidden = availableOfflineHidden
            self?.downloadProgressView?.isHidden = downloadProgressHidden
            self?.downloadProgressView?.updateProgress(progress)
        }

        configureForSelection()
    }

    func configureWith(driveFileManager: DriveFileManager, file: File, selectionMode: Bool = false) {
        configure(with: FileViewModel(driveFileManager: driveFileManager, file: file, selectionMode: selectionMode))
    }

    /// Update the cell selection mode.
    /// - Parameter selectionMode: The new selection mode (enabled/disabled).
    func setSelectionMode(_ selectionMode: Bool) {
        guard viewModel != nil else { return }
        viewModel.selectionMode = selectionMode
        configure(with: viewModel)
    }

    private func configureForSelection() {
        guard viewModel?.selectionMode == true else { return }
        checkmarkImage?.image = isSelected ? KDriveResourcesAsset.select.image : FileCollectionViewCell.emptyCheckmarkImage
        checkmarkImage?.isAccessibilityElement = true
        checkmarkImage?.accessibilityLabel = isSelected ? KDriveResourcesStrings.Localizable.contentDescriptionIsSelected : ""
        checkmarkImage?.backgroundColor = nil
        checkmarkImage?.contentMode = .scaleAspectFit
        checkmarkImage?.layer.cornerRadius = 0
        checkmarkImage?.layer.masksToBounds = false
    }

    func configureLoading() {
        titleLabel.text = " "
        let titleLayer = CALayer()
        titleLayer.anchorPoint = .zero
        titleLayer.bounds = CGRect(x: 0, y: 0, width: 100, height: 15)
        titleLayer.backgroundColor = KDriveResourcesAsset.loaderDarkerDefaultColor.color.cgColor
        titleLabel.layer.addSublayer(titleLayer)
        favoriteImageView?.isHidden = true
        logoImage.image = nil
        logoImage.backgroundColor = KDriveResourcesAsset.loaderDarkerDefaultColor.color
        moreButton.isHidden = true
        checkmarkImage?.isHidden = true
    }

    @IBAction func moreButtonTap(_ sender: Any) {
        delegate?.didTapMoreButton(self)
    }
}

extension FileCollectionViewCell: UICollectionViewDelegate, UICollectionViewDataSource, UICollectionViewDelegateFlowLayout {
    func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int {
        return min(viewModel.categories.count, 3)
    }

    func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        let cell = collectionView.dequeueReusableCell(type: CategoryBadgeCollectionViewCell.self, for: indexPath)
        let category = viewModel.categories[indexPath.row]
        let more = indexPath.item == 2 && viewModel.categories.count > 3 ? viewModel.categories.count - 3 : nil
        cell.configure(with: category, more: more)
        return cell
    }

    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, sizeForItemAt indexPath: IndexPath) -> CGSize {
        return CGSize(width: 16, height: 16)
    }
}
